@@ -1,7 +1,7 @@
 /*
  * <security/_pam_types.h>
  *
- * $Id: _pam_types.h,v 1.3 2000/11/19 23:54:02 agmorgan Exp $
+ * $Id: _pam_types.h,v 1.4 2001/01/22 06:07:29 agmorgan Exp $
  *
  * This file defines all of the types common to the Linux-PAM library
  * applications and modules.
@@ -14,6 +14,11 @@
 #ifndef _SECURITY__PAM_TYPES_H
 #define _SECURITY__PAM_TYPES_H
 
+#ifndef __LIBPAM_VERSION
+# define __LIBPAM_VERSION __libpam_version
+#endif
+extern unsigned int __libpam_version;
+
 /*
  * include local definition for POSIX - NULL
  */
@@ -86,7 +91,10 @@ typedef struct pam_handle pam_handle_t;
 				   calling again, verify that conversation
 				   is completed */
 
-/* Add new #define's here */
+/*
+ * Add new #define's here - take care to also extend the libpam code:
+ * pam_strerror() and "libpam/pam_tokens.h" .
+ */
 
 #define _PAM_RETURN_VALUES 32   /* this is the number of return values */
 
