@@ -1,5 +1,5 @@
 /*
- * $Id: misc_conv.c,v 1.3 2001/01/20 22:29:47 agmorgan Exp $
+ * $Id: misc_conv.c,v 1.5 2001/10/11 04:14:30 agmorgan Exp $
  *
  * A generic conversation function for text based applications
  *
@@ -57,7 +57,7 @@ void (*pam_binary_handler_free)(void *appdata, pamc_bp_t *prompt_p)
 
 /* the following code is used to get text input */
 
-volatile static int expired=0;
+static volatile int expired=0;
 
 /* return to the previous signal handling */
 static void reset_alarm(struct sigaction *o_ptr)
@@ -130,12 +130,15 @@ static int get_delay(void)
 static char *read_string(int echo, const char *prompt)
 {
     struct termios term_before, term_tmp;
-    char line[INPUTSIZE];
+    char line[INPUTSIZE], *input;
     struct sigaction old_sig;
     int delay, nc, have_term=0;
-
+    sigset_t oset, nset;
+ 
     D(("called with echo='%s', prompt='%s'.", echo ? "ON":"OFF" , prompt));
 
+    input = line;
+
     if (isatty(STDIN_FILENO)) {                      /* terminal state */
 
 	/* is a terminal so record settings and flush it */
@@ -149,6 +152,16 @@ static char *read_string(int echo, const char *prompt)
 	}
 	have_term = 1;
 
+	/*
+	 * We make a simple attempt to block TTY signals from terminating
+	 * the conversation without giving PAM a chance to clean up.
+	 */
+
+	sigemptyset(&nset); 
+	sigaddset(&nset, SIGINT); 
+	sigaddset(&nset, SIGTSTP); 
+	(void) sigprocmask(SIG_BLOCK, &nset, &oset);
+
     } else if (!echo) {
 	D(("<warning: cannot turn echo off>"));
     }
@@ -180,35 +193,67 @@ static char *read_string(int echo, const char *prompt)
 	    if (expired) {
 		delay = get_delay();
 	    } else if (nc > 0) {                 /* we got some user input */
-		char *input;
+		D(("we got some user input"));
 
 		if (nc > 0 && line[nc-1] == '\n') {     /* <NUL> terminate */
 		    line[--nc] = '\0';
 		} else {
+		    if (echo) {
+			fprintf(stderr, "\n");
+		    }
 		    line[nc] = '\0';
 		}
 		input = x_strdup(line);
 		_pam_overwrite(line);
 
-		return input;                  /* return malloc()ed string */
+		goto cleanexit;                /* return malloc()ed string */
+
 	    } else if (nc == 0) {                                /* Ctrl-D */
 		D(("user did not want to type anything"));
-		fprintf(stderr, "\n");
-		break;
+
+		input = x_strdup("");
+		if (echo) {
+		    fprintf(stderr, "\n");
+		}
+		goto cleanexit;                /* return malloc()ed "" */
 	    }
 	}
     }
 
     /* getting here implies that the timer expired */
-    if (have_term)
-	(void) tcsetattr(STDIN_FILENO, TCSADRAIN, &term_before);
 
-    memset(line, 0, INPUTSIZE);                      /* clean up */
-    return NULL;
+    D(("the timer appears to have expired"));
+
+    input = NULL;
+    _pam_overwrite(line);
+
+ cleanexit:
+
+    if (have_term) {
+	(void) sigprocmask(SIG_SETMASK, &oset, NULL);
+	(void) tcsetattr(STDIN_FILENO, TCSADRAIN, &term_before);
+    }
+
+    D(("returning [%s]", input));
+
+    return input;
 }
 
 /* end of read_string functions */
 
+/*
+ * This conversation function is supposed to be a generic PAM one.
+ * Unfortunately, it is _not_ completely compatible with the Solaris PAM
+ * codebase.
+ *
+ * Namely, for msgm's that contain multiple prompts, this function
+ * interprets "const struct pam_message **msgm" as equivalent to
+ * "const struct pam_message *msgm[]". The Solaris module
+ * implementation interprets the **msgm object as a pointer to a
+ * pointer to an array of "struct pam_message" objects (that is, a
+ * confusing amount of pointer indirection).
+ */
+
 int misc_conv(int num_msg, const struct pam_message **msgm,
 	      struct pam_response **response, void *appdata_ptr)
 {
@@ -294,8 +339,6 @@ int misc_conv(int num_msg, const struct pam_message **msgm,
 	}
     }
 
-    /* New (0.59+) behavior is to always have a reply - this is
-       compatable with the X/Open (March 1997) spec. */
     *response = reply;
     reply = NULL;
 
@@ -303,6 +346,8 @@ int misc_conv(int num_msg, const struct pam_message **msgm,
 
 failed_conversation:
 
+    D(("the conversation failed"));
+
     if (reply) {
 	for (count=0; count<num_msg; ++count) {
 	    if (reply[count].resp == NULL) {
