@@ -1,13 +1,13 @@
 /* pam_group module */
 
 /*
- * $Id: pam_group.c,v 1.3 2000/11/26 07:32:39 agmorgan Exp $
+ * $Id: pam_group.c,v 1.4 2001/05/08 05:02:50 agmorgan Exp $
  *
  * Written by Andrew Morgan <morgan@linux.kernel.org> 1996/7/6
  */
 
 const static char rcsid[] =
-"$Id: pam_group.c,v 1.3 2000/11/26 07:32:39 agmorgan Exp $;\n"
+"$Id: pam_group.c,v 1.4 2001/05/08 05:02:50 agmorgan Exp $;\n"
 "Version 0.5 for Linux-PAM\n"
 "Copyright (c) Andrew G. Morgan 1996 <morgan@linux.kernel.org>\n";
 
@@ -36,6 +36,13 @@ const static char rcsid[] =
 #define PAM_GROUP_BUFLEN        1000
 #define FIELD_SEPARATOR         ';'   /* this is new as of .02 */
 
+#ifdef TRUE 
+# undef TRUE 
+#endif 
+#ifdef FALSE 
+# undef FALSE 
+#endif
+
 typedef enum { FALSE, TRUE } boolean;
 typedef enum { AND, OR } operator;
 
