@@ -2,7 +2,7 @@
  * pam_limits - impose resource limits when opening a user session
  *
  * 1.6 - modified for PLD (added process priority settings) 
- *       by Marcin Korzonek <mkorz@shadow.eu.org
+ *       by Marcin Korzonek <mkorz@shadow.eu.org>
  * 1.5 - Elliot Lee's "max system logins patch"
  * 1.4 - addressed bug in configuration file parser
  * 1.3 - modified the configuration file format
@@ -24,6 +24,7 @@
 #include <string.h>
 #include <ctype.h>
 #include <stdlib.h>
+#include <errno.h>
 #include <syslog.h>
 #include <stdarg.h>
 #include <sys/types.h>
@@ -45,6 +46,16 @@
 #define LIMITS_DEF_GROUP    1 /* limit was set by a group entry */
 #define LIMITS_DEF_DEFAULT  2 /* limit was set by an default entry */
 #define LIMITS_DEF_NONE     3 /* this limit was not set yet */
+#define LIMITS_DEF_ALL      4 /* limit was set by an default entry */
+#define LIMITS_DEF_ALLGROUP 5 /* limit was set by a group entry */
+
+static const char *limits_def_names[] = {
+       "USER",
+       "GROUP",
+       "DEFAULT",
+       "NONE",
+       NULL,
+};
 
 struct user_limits_struct {
     int src_soft;
@@ -59,8 +70,10 @@ struct pam_limit_s {
     int flag_numsyslogins; /* whether to limit logins only for a
 			      specific user or to count all logins */
     int priority;	 /* the priority to run user process with */
+    int supported[RLIM_NLIMITS];
     struct user_limits_struct limits[RLIM_NLIMITS];
     char conf_file[BUFSIZ];
+    int utmp_after_pam_call;
 };
 
 #define LIMIT_LOGIN RLIM_NLIMITS+1
@@ -92,28 +105,32 @@ static void _pam_log(int err, const char *format, ...)
 
 #define PAM_DEBUG_ARG       0x0001
 #define PAM_DO_SETREUID     0x0002
+#define PAM_UTMP_EARLY      0x0004
 
 static int _pam_parse(int argc, const char **argv, struct pam_limit_s *pl)
 {
-     int ctrl=0;
+    int ctrl=0;
 
-     /* step through arguments */
-     for (ctrl=0; argc-- > 0; ++argv) {
+    /* step through arguments */
+    for (ctrl=0; argc-- > 0; ++argv) {
 
-          /* generic options */
+	/* generic options */
 
-          if (!strcmp(*argv,"debug"))
-               ctrl |= PAM_DEBUG_ARG;
-          else if (!strncmp(*argv,"conf=",5))
-                strcpy(pl->conf_file,*argv+5);
-	  else if (!strncmp(*argv,"change_uid",10))
-		ctrl |= PAM_DO_SETREUID;
-          else {
-               _pam_log(LOG_ERR,"pam_parse: unknown option; %s",*argv);
-          }
-     }
+	if (!strcmp(*argv,"debug")) {
+	    ctrl |= PAM_DEBUG_ARG;
+	} else if (!strncmp(*argv,"conf=",5)) {
+	    strncpy(pl->conf_file,*argv+5,sizeof(pl->conf_file)-1);
+	} else if (!strncmp(*argv,"change_uid",10)) {
+	    ctrl |= PAM_DO_SETREUID;
+	} else if (!strcmp(*argv,"utmp_early")) {
+	    ctrl |= PAM_UTMP_EARLY;
+	} else {
+	    _pam_log(LOG_ERR,"pam_parse: unknown option; %s",*argv);
+	}
+    }
+    pl->conf_file[sizeof(pl->conf_file) - 1] = '\0';
 
-     return ctrl;
+    return ctrl;
 }
 
 
@@ -124,54 +141,9 @@ static int _pam_parse(int argc, const char **argv, struct pam_limit_s *pl)
 # define LIMITS_FILE "/etc/security/limits.conf"
 #endif
 
-#define LIMIT_ERR 1 /* error setting a limit */
-#define LOGIN_ERR 2 /* too many logins err */
-
-/* Counts the number of user logins and check against the limit*/
-static int check_logins(const char *name, int limit, int ctrl,
-			struct pam_limit_s *pl)
-{
-    struct utmp *ut;
-    unsigned int count;
-
-    if (ctrl & PAM_DEBUG_ARG) {
-	 _pam_log(LOG_DEBUG, "checking logins for '%s' / %d\n", name,limit);
-    }
-
-    if (limit < 0)
-        return 0; /* no limits imposed */
-    if (limit == 0) /* maximum 0 logins ? */ {
-        _pam_log(LOG_WARNING, "No logins allowed for '%s'\n", name);
-        return LOGIN_ERR;
-    }
-
-    setutent();
-    count = 0;
-    while((ut = getutent())) {
-#ifdef USER_PROCESS
-        if (ut->ut_type != USER_PROCESS)
-            continue;
-#endif
-        if (ut->UT_USER[0] == '\0')
-            continue;
-        if (!pl->flag_numsyslogins
-	    && strncmp(name, ut->UT_USER, sizeof(ut->UT_USER)) != 0)
-            continue;
-        if (++count >= limit)
-            break;
-    }
-    endutent();
-    if (count >= limit) {
-	if (name) {
-	    _pam_log(LOG_WARNING, "Too many logins (max %d) for %s",
-		     limit, name);
-	} else {
-	    _pam_log(LOG_WARNING, "Too many system logins (max %d)", limit);
-	}
-        return LOGIN_ERR;
-    }
-    return 0;
-}
+#define LIMITED_OK 0 /* limit setting appeared to work */
+#define LIMIT_ERR  1 /* error setting a limit */
+#define LOGIN_ERR  2 /* too many logins err */
 
 /* checks if a user is on a list of members of the GID 0 group */
 static int is_on_list(char * const *list, const char *member)
@@ -184,32 +156,31 @@ static int is_on_list(char * const *list, const char *member)
     return 0;
 }
 
-/* Checks if a user is a member of a group */
-static int is_on_group(const char *user_name, const char *group_name)
+/*
+ * Checks if a user is a member of a group - return non-zero if
+ * the user is in the group.
+ */
+static int is_in_group(const char *user_name, const char *group_name)
 {
     struct passwd *pwd;
     struct group *grp, *pgrp;
     char uname[LINE_LENGTH], gname[LINE_LENGTH];
     
-    if (!strlen(user_name))
+    if (!user_name || !strlen(user_name))
         return 0;
-    if (!strlen(group_name))
+    if (!group_name || !strlen(group_name))
         return 0;
     memset(uname, 0, sizeof(uname));
-    strncpy(uname, user_name, LINE_LENGTH);
+    strncpy(uname, user_name, sizeof(uname)-1);
     memset(gname, 0, sizeof(gname));
-    strncpy(gname, group_name, LINE_LENGTH);
+    strncpy(gname, group_name, sizeof(gname)-1);
         
-    setpwent();
     pwd = getpwnam(uname);
-    endpwent();
     if (!pwd)
         return 0;
 
     /* the info about this group */
-    setgrent();
     grp = getgrnam(gname);
-    endgrent();
     if (!grp)
         return 0;
     
@@ -218,9 +189,7 @@ static int is_on_group(const char *user_name, const char *group_name)
         return 1;
 
     /* next check: user primary group is group_name ? */
-    setgrent();
     pgrp = getgrgid(pwd->pw_gid);
-    endgrent();
     if (!pgrp)
         return 0;
     if (!strcmp(pgrp->gr_name, gname))
@@ -229,6 +198,81 @@ static int is_on_group(const char *user_name, const char *group_name)
     return 0;
 }
     
+/* Counts the number of user logins and check against the limit*/
+static int check_logins(const char *name, int limit, int ctrl,
+			struct pam_limit_s *pl)
+{
+    struct utmp *ut;
+    unsigned int count;
+
+    if (ctrl & PAM_DEBUG_ARG) {
+        _pam_log(LOG_DEBUG, "checking logins for '%s' (maximum of %d)\n",
+                 name, limit);
+    }
+
+    if (limit < 0)
+        return 0; /* no limits imposed */
+    if (limit == 0) /* maximum 0 logins ? */ {
+        _pam_log(LOG_WARNING, "No logins allowed for '%s'\n", name);
+        return LOGIN_ERR;
+    }
+
+    setutent();
+
+    /* Because there is no definition about when an application
+       actually adds a utmp entry, some applications bizarrely do the
+       utmp call before the have PAM authenticate them to the system:
+       you're logged it, sort of...? Anyway, you can use the
+       "utmp_early" module argument in your PAM config file to make
+       allowances for this sort of problem. (There should be a PAM
+       standard for this, since if a module wants to actually map a
+       username then any early utmp entry will be for the unmapped
+       name = broken.) */
+    
+    if (ctrl & PAM_UTMP_EARLY) {
+	count = 0;
+    } else {
+	count = 1;
+    }
+
+    while((ut = getutent())) {
+#ifdef USER_PROCESS
+        if (ut->ut_type != USER_PROCESS) {
+            continue;
+	}
+#endif
+        if (ut->UT_USER[0] == '\0') {
+            continue;
+	}
+        if (!pl->flag_numsyslogins) {
+	    if (((pl->login_limit_def == LIMITS_DEF_USER)
+	         || (pl->login_limit_def == LIMITS_DEF_GROUP)
+		 || (pl->login_limit_def == LIMITS_DEF_DEFAULT))
+		&& strncmp(name, ut->UT_USER, sizeof(ut->UT_USER)) != 0) {
+                continue;
+	    }
+	    if ((pl->login_limit_def == LIMITS_DEF_ALLGROUP)
+		&& !is_in_group(ut->UT_USER, name)) {
+                continue;
+	    }
+	}
+	if (++count > limit) {
+	    break;
+	}
+    }
+    endutent();
+    if (count > limit) {
+	if (name) {
+	    _pam_log(LOG_WARNING, "Too many logins (max %d) for %s",
+		     limit, name);
+	} else {
+	    _pam_log(LOG_WARNING, "Too many system logins (max %d)", limit);
+	}
+        return LOGIN_ERR;
+    }
+    return 0;
+}
+
 static int init_limits(struct pam_limit_s *pl)
 {
     int i;
@@ -236,38 +280,20 @@ static int init_limits(struct pam_limit_s *pl)
 
     D(("called."));
 
-    for(i = 0; i < RLIM_NLIMITS; i++) 
-	retval |= getrlimit(i, &pl->limits[i].limit);
-    
-    pl->limits[RLIMIT_CPU].src_soft = LIMITS_DEF_NONE;
-    pl->limits[RLIMIT_CPU].src_hard = LIMITS_DEF_NONE;
-
-    pl->limits[RLIMIT_FSIZE].src_soft = LIMITS_DEF_NONE;
-    pl->limits[RLIMIT_FSIZE].src_hard = LIMITS_DEF_NONE;
-
-    pl->limits[RLIMIT_DATA].src_soft = LIMITS_DEF_NONE;
-    pl->limits[RLIMIT_DATA].src_hard = LIMITS_DEF_NONE;
-
-    pl->limits[RLIMIT_STACK].src_soft = LIMITS_DEF_NONE;
-    pl->limits[RLIMIT_STACK].src_hard = LIMITS_DEF_NONE;
-
-    pl->limits[RLIMIT_CORE].src_soft = LIMITS_DEF_NONE;
-    pl->limits[RLIMIT_CORE].src_hard = LIMITS_DEF_NONE;
-
-    pl->limits[RLIMIT_RSS].src_soft = LIMITS_DEF_NONE;
-    pl->limits[RLIMIT_RSS].src_hard = LIMITS_DEF_NONE;
-
-    pl->limits[RLIMIT_NPROC].src_soft = LIMITS_DEF_NONE;
-    pl->limits[RLIMIT_NPROC].src_hard = LIMITS_DEF_NONE;
-
-    pl->limits[RLIMIT_NOFILE].src_soft = LIMITS_DEF_NONE;
-    pl->limits[RLIMIT_NOFILE].src_hard = LIMITS_DEF_NONE;
-
-    pl->limits[RLIMIT_MEMLOCK].src_soft = LIMITS_DEF_NONE;
-    pl->limits[RLIMIT_MEMLOCK].src_hard = LIMITS_DEF_NONE;
-
-    pl->limits[RLIMIT_AS].src_soft = LIMITS_DEF_NONE;
-    pl->limits[RLIMIT_AS].src_hard = LIMITS_DEF_NONE;
+    for(i = 0; i < RLIM_NLIMITS; i++) {
+	int r = getrlimit(i, &pl->limits[i].limit);
+	if (r == -1) {
+	    if (errno == EINVAL) {
+		pl->supported[i] = 0;
+	    } else {
+		retval = !PAM_SUCCESS;
+	    }
+	} else {
+	    pl->supported[i] = 1;
+	    pl->limits[i].src_soft = LIMITS_DEF_NONE;
+	    pl->limits[i].src_hard = LIMITS_DEF_NONE;
+	}
+    }
 
     pl->priority = 0;
     pl->login_limit = -2;
@@ -287,8 +313,9 @@ static void process_limit(int source, const char *lim_type,
     const char *value_orig = lim_value;
         
     if (ctrl & PAM_DEBUG_ARG)
-	 _pam_log(LOG_DEBUG, "%s: processing(%d) %s %s %s\n",
-		  __FUNCTION__,source,lim_type,lim_item,lim_value);
+	 _pam_log(LOG_DEBUG, "%s: processing %s %s %s for %s\n",
+		  __FUNCTION__,lim_type,lim_item,lim_value,
+		  limits_def_names[source]);
 
     if (strcmp(lim_item, "cpu") == 0)
         limit_item = RLIMIT_CPU;
@@ -310,6 +337,10 @@ static void process_limit(int source, const char *lim_type,
 	limit_item = RLIMIT_MEMLOCK;
     else if (strcmp(lim_item, "as") == 0)
 	limit_item = RLIMIT_AS;
+#ifdef RLIMIT_LOCKS
+    else if (strcmp(lim_item, "locks") == 0)
+	limit_item = RLIMIT_LOCKS;
+#endif
     else if (strcmp(lim_item, "maxlogins") == 0) {
 	limit_item = LIMIT_LOGIN;
 	pl->flag_numsyslogins = 0;
@@ -333,8 +364,14 @@ static void process_limit(int source, const char *lim_type,
         _pam_log(LOG_DEBUG,"unknown limit type '%s'", lim_type);
         return;
     }
-    
+
+    /*
+     * there is a warning here because the library prototype for this
+     * function is incorrect.
+     */
     limit_value = strtol(lim_value, endptr, 10);
+
+    /* special case value when limiting logins */
     if (limit_value == 0 && value_orig == *endptr) { /* no chars read */
         if (strcmp(lim_value,"-") != 0) {
             _pam_log(LOG_DEBUG,"wrong limit value '%s'", lim_value);
@@ -348,7 +385,16 @@ static void process_limit(int source, const char *lim_type,
             } else
                 limit_value = -1;
     }
-    
+
+    /* one more special case when limiting logins */
+    if ((source == LIMITS_DEF_ALL || source == LIMITS_DEF_ALLGROUP)
+		&& (limit_item != LIMIT_LOGIN)) {
+	if (ctrl & PAM_DEBUG_ARG)
+	    _pam_log(LOG_DEBUG,
+			"'%%' domain valid for maxlogins type only");
+	return;
+    }
+
     switch(limit_item) {
         case RLIMIT_CPU:
             limit_value *= 60;
@@ -363,10 +409,10 @@ static void process_limit(int source, const char *lim_type,
             limit_value *= 1024;
             break;
     }
-    
-    if (limit_item != LIMIT_LOGIN && limit_item != LIMIT_NUMSYSLOGINS 
-		    && limit_item != LIMIT_PRI
-		    ) {
+
+    if ( (limit_item != LIMIT_LOGIN)
+	 && (limit_item != LIMIT_NUMSYSLOGINS)
+	 && (limit_item != LIMIT_PRI) ) {
         if (limit_type & LIMIT_SOFT) {
 	    if (pl->limits[limit_item].src_soft < source) {
                 return;
@@ -383,10 +429,11 @@ static void process_limit(int source, const char *lim_type,
                 pl->limits[limit_item].src_hard = source;
             }
 	}
-    } else 
+    } else {
+	/* recent kernels support negative priority limits (=raise priority) */
+
 	if (limit_item == LIMIT_PRI) {
-		/* additional check */
-		pl->priority = ((limit_value>0)?limit_value:0);
+		pl->priority = limit_value;
 	} else {
 	        if (pl->login_limit_def < source) {
 	            return;
@@ -395,6 +442,7 @@ static void process_limit(int source, const char *lim_type,
 	            pl->login_limit_def = source;
         	}
 	}
+    }
     return;
 }
 
@@ -430,7 +478,8 @@ static int parse_config_file(const char *uname, int ctrl,
         /* skip the leading white space */
         while (*tptr && isspace(*tptr))
             tptr++;
-        strcpy(buf, (const char *)tptr);
+        strncpy(buf, tptr, sizeof(buf)-1);
+	buf[sizeof(buf)-1] = '\0';
                                 
         /* Rip off the comments */
         tptr = strchr(buf,'#');
@@ -468,20 +517,39 @@ static int parse_config_file(const char *uname, int ctrl,
             if (strcmp(uname, domain) == 0) /* this user have a limit */
                 process_limit(LIMITS_DEF_USER, ltype, item, value, ctrl, pl);
             else if (domain[0]=='@') {
-                if (is_on_group(uname, domain+1))
+		    if (ctrl & PAM_DEBUG_ARG) {
+			_pam_log(LOG_DEBUG, "checking if %s is in group %s",
+			 	uname, domain + 1);
+		    }
+                if (is_in_group(uname, domain+1))
                     process_limit(LIMITS_DEF_GROUP, ltype, item, value, ctrl,
 				  pl);
+            } else if (domain[0]=='%') {
+		    if (ctrl & PAM_DEBUG_ARG) {
+			_pam_log(LOG_DEBUG, "checking if %s is in group %s",
+			 	uname, domain + 1);
+		    }
+		if (strcmp(domain,"%") == 0)
+		    process_limit(LIMITS_DEF_ALL, ltype, item, value, ctrl,
+				  pl);
+		else if (is_in_group(uname, domain+1))
+                    process_limit(LIMITS_DEF_ALLGROUP, ltype, item, value, ctrl,
+				  pl);
             } else if (strcmp(domain, "*") == 0)
                 process_limit(LIMITS_DEF_DEFAULT, ltype, item, value, ctrl,
 			      pl);
 	} else if (i == 2 && ltype[0] == '-') { /* Probably a no-limit line */
 	    if (strcmp(uname, domain) == 0) {
-		_pam_log(LOG_DEBUG, "no limits for '%s'", uname);
+		if (ctrl & PAM_DEBUG_ARG) {
+		    _pam_log(LOG_DEBUG, "no limits for '%s'", uname);
+		}
 		fclose(fil);
 		return PAM_IGNORE;
-	    } else if (domain[0] == '@' && is_on_group(uname, domain+1)) {
-		_pam_log(LOG_DEBUG, "no limits for '%s' in group '%s'",
-			 uname, domain+1);
+	    } else if (domain[0] == '@' && is_in_group(uname, domain+1)) {
+		if (ctrl & PAM_DEBUG_ARG) {
+		    _pam_log(LOG_DEBUG, "no limits for '%s' in group '%s'",
+			     uname, domain+1);
+		}
 		fclose(fil);
 		return PAM_IGNORE;
 	    }
@@ -493,30 +561,53 @@ static int parse_config_file(const char *uname, int ctrl,
     return PAM_SUCCESS;    
 }
 
-static int setup_limits(const char * uname, int ctrl, struct pam_limit_s *pl)
+static int setup_limits(const char * uname, uid_t uid, int ctrl,
+			struct pam_limit_s *pl)
 {
     int i;
-    int retval = PAM_SUCCESS;
-    
-    for (i=0; i<RLIM_NLIMITS; i++) {
+    int status;
+    int retval = LIMITED_OK;
+
+    if (uid == 0) {
+	/* do not impose limits (+ve limits anyway) on the superuser */
+	if (pl->priority > 0) {
+	    if (ctrl & PAM_DEBUG_ARG) {
+		_pam_log(LOG_DEBUG, "user '%s' has UID 0 - no limits imposed",
+			 uname);
+	    }
+            pl->priority = 0;
+	}
+    }
+
+    for (i=0, status=LIMITED_OK; i<RLIM_NLIMITS; i++) {
         if (pl->limits[i].limit.rlim_cur > pl->limits[i].limit.rlim_max)
             pl->limits[i].limit.rlim_cur = pl->limits[i].limit.rlim_max;
-        retval |= setrlimit(i, &pl->limits[i].limit);
+	if (!pl->supported[i]) {
+	    /* skip it if its not known to the system */
+	    continue;
+	}
+	status |= setrlimit(i, &pl->limits[i].limit);
     }
     
-    if (retval != PAM_SUCCESS)
+    if (status) {
         retval = LIMIT_ERR;
+    }
 
-    retval=setpriority(PRIO_PROCESS, 0, pl->priority);
-    
-    if (retval != PAM_SUCCESS)
+    status = setpriority(PRIO_PROCESS, 0, pl->priority);
+    if (status != 0) {
         retval = LIMIT_ERR;
+    }
 
-    if (pl->login_limit > 0) {
-        if (check_logins(uname, pl->login_limit, ctrl, pl) == LOGIN_ERR)
+    if (uid == 0) {
+	D(("skip login limit check for uid=0"));
+    } else if (pl->login_limit > 0) {
+        if (check_logins(uname, pl->login_limit, ctrl, pl) == LOGIN_ERR) {
             retval |= LOGIN_ERR;
-    } else if (pl->login_limit == 0)
+	}
+    } else if (pl->login_limit == 0) {
         retval |= LOGIN_ERR;
+    }
+
     return retval;
 }
             
@@ -541,9 +632,7 @@ PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh, int flags,
         return PAM_SESSION_ERR;
      }
 	
-    setpwent();
     pwd = getpwnam(user_name);
-    endpwent();
     if (!pwd) {
         if (ctrl & PAM_DEBUG_ARG)
             _pam_log(LOG_WARNING, "open_session username '%s' does not exist",
@@ -551,14 +640,6 @@ PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh, int flags,
         return PAM_SESSION_ERR;
     }
                      
-    /* do not impose limits on UID 0 accounts */
-    if (!pwd->pw_uid) {
-        if (ctrl & PAM_DEBUG_ARG)
-            _pam_log(LOG_DEBUG, "user '%s' have UID 0 - no limits imposed",
-                                user_name);
-        return PAM_SUCCESS;
-    }
-        
     retval = init_limits(&pl);
     if (retval != PAM_SUCCESS) {
         _pam_log(LOG_WARNING, "cannot initialize");
@@ -575,12 +656,11 @@ PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh, int flags,
         return PAM_IGNORE;
     }
 
-    if (ctrl & PAM_DO_SETREUID)
+    if (ctrl & PAM_DO_SETREUID) {
 	setreuid(pwd->pw_uid, -1);
-    retval = setup_limits(pwd->pw_name, ctrl, &pl);
-    if (retval & LOGIN_ERR) {
-        printf("\nToo many logins for '%s'\n",pwd->pw_name);
-        sleep(2);
+    }
+    retval = setup_limits(pwd->pw_name, pwd->pw_uid, ctrl, &pl);
+    if (retval != LIMITED_OK) {
         return PAM_PERM_DENIED;
     }
 
