@@ -1,21 +1,13 @@
 /* pam_mail module */
 
 /*
- * $Id: pam_mail.c,v 1.2 2000/11/19 23:54:04 agmorgan Exp $
+ * $Id: pam_mail.c,v 1.3 2000/12/04 20:59:13 baggins Exp $
  *
  * Written by Andrew Morgan <morgan@linux.kernel.org> 1996/3/11
  * $HOME additions by David Kinchlea <kinch@kinch.ark.com> 1997/1/7
  * mailhash additions by Chris Adams <cadams@ro.com> 1998/7/11
  */
 
-#define DEFAULT_MAIL_DIRECTORY    "/var/spool/mail"
-#define MAIL_FILE_FORMAT          "%s%s/%s"
-#define MAIL_ENV_NAME             "MAIL"
-#define MAIL_ENV_FORMAT           MAIL_ENV_NAME "=%s"
-#define YOUR_MAIL_VERBOSE_FORMAT  "You have %s mail in %s."
-#define YOUR_MAIL_STANDARD_FORMAT "You have %smail."
-#define NO_MAIL_STANDARD_FORMAT   "No mail." 
-
 #include <security/_pam_aconf.h>
 
 #include <ctype.h>
@@ -34,6 +26,14 @@
 #include <pwdb/pwdb_public.h>
 #endif
 
+#define DEFAULT_MAIL_DIRECTORY    PAM_PATH_MAILDIR
+#define MAIL_FILE_FORMAT          "%s%s/%s"
+#define MAIL_ENV_NAME             "MAIL"
+#define MAIL_ENV_FORMAT           MAIL_ENV_NAME "=%s"
+#define YOUR_MAIL_VERBOSE_FORMAT  "You have %s mail in %s."
+#define YOUR_MAIL_STANDARD_FORMAT "You have %smail."
+#define NO_MAIL_STANDARD_FORMAT   "No mail." 
+
 /*
  * here, we make a definition for the externally accessible function
  * in this file (this definition is required for static a module
