@@ -1,11 +1,7 @@
 #ifndef MD5_H
 #define MD5_H
 
-#ifdef __alpha
 typedef unsigned int uint32;
-#else
-typedef unsigned long uint32;
-#endif
 
 struct MD5Context {
     uint32 buf[4];
