@@ -1,18 +1,18 @@
 /* pam_stress module */
 
-/* $Id: pam_stress.c,v 1.2 1999/07/08 05:02:06 morgan Exp $
+/* $Id: pam_stress.c,v 1.2 2000/11/19 23:54:05 agmorgan Exp $
  *
  * created by Andrew Morgan <morgan@linux.kernel.org> 1996/3/12
  */
 
-#ifdef linux
-# define _GNU_SOURCE
-# include <features.h>
-#endif
+#include <security/_pam_aconf.h>
 
 #include <stdlib.h>
 #include <stdio.h>
+
+#define __USE_BSD
 #include <syslog.h>
+
 #include <stdarg.h>
 #include <string.h>
 #include <unistd.h>
