@@ -1,7 +1,7 @@
 /* pam_time module */
 
 /*
- * $Id: pam_time.c,v 1.3 2000/11/26 07:32:39 agmorgan Exp $
+ * $Id: pam_time.c,v 1.4 2001/05/08 05:02:50 agmorgan Exp $
  *
  * Written by Andrew Morgan <morgan@linux.kernel.org> 1996/6/22
  * (File syntax and much other inspiration from the shadow package
@@ -9,7 +9,7 @@
  */
 
 const static char rcsid[] =
-"$Id: pam_time.c,v 1.3 2000/11/26 07:32:39 agmorgan Exp $;\n"
+"$Id: pam_time.c,v 1.4 2001/05/08 05:02:50 agmorgan Exp $;\n"
 "\t\tVersion 0.22 for Linux-PAM\n"
 "Copyright (C) Andrew G. Morgan 1996 <morgan@linux.kernel.org>\n";
 
@@ -36,6 +36,13 @@ const static char rcsid[] =
 #define PAM_TIME_BUFLEN        1000
 #define FIELD_SEPARATOR        ';'   /* this is new as of .02 */
 
+#ifdef TRUE 
+# undef TRUE 
+#endif 
+#ifdef FALSE 
+# undef FALSE 
+#endif
+
 typedef enum { FALSE, TRUE } boolean;
 typedef enum { AND, OR } operator;
 
