@@ -1,5 +1,5 @@
 /* 
- * $Id: support.c,v 1.7 2000/12/20 05:15:05 vorlon Exp $
+ * $Id: support.c,v 1.13 2002/07/09 04:44:18 agmorgan Exp $
  *
  * Copyright information at end of file.
  */
@@ -15,6 +15,7 @@
 #include <shadow.h>
 #include <limits.h>
 #include <utmp.h>
+#include <errno.h>
 
 #include <security/_pam_macros.h>
 #include <security/pam_modules.h>
@@ -122,9 +123,10 @@ char *PAM_getlogin(void)
 		D(("PAM_getlogin ttyname: %s", curr_tty));
 		curr_tty += 5;
 		setutent();
-		strncpy(line.ut_line, curr_tty, sizeof line.ut_line);
+		strncpy(line.ut_line, curr_tty, sizeof(line.ut_line));
 		if ((ut = getutline(&line)) != NULL) {
 			strncpy(curr_user, ut->ut_user, sizeof(ut->ut_user));
+			curr_user[sizeof(curr_user) - 1] = '\0';
 			retval = curr_user;
 		}
 		endutent();
@@ -314,6 +316,13 @@ int _unix_blankpasswd(unsigned int ctrl, const char *name)
 	struct spwd *spwdent = NULL;
 	char *salt = NULL;
 	int retval;
+#if HAVE_GETPWNAM_R
+	char *buf = NULL;
+	int bufsize = 0;
+	struct passwd pwd_buf;
+
+	pwd = &pwd_buf;
+#endif
 
 	D(("called"));
 
@@ -327,7 +336,25 @@ int _unix_blankpasswd(unsigned int ctrl, const char *name)
 		return 0;	/* will fail but don't let on yet */
 
 	/* UNIX passwords area */
-	pwd = getpwnam(name);	/* Get password file entry... */
+
+	/* Get password file entry... */
+#if HAVE_GETPWNAM_R
+	bufsize = 1024;
+	buf = malloc(bufsize);
+
+	if ((retval = getpwnam_r(name, pwd, buf, bufsize, &pwd))) {
+		pwd = NULL;
+	}
+	while (retval == ERANGE) {
+		bufsize += 1024;
+		buf = realloc(buf, bufsize);
+		if ((retval = getpwnam_r(name, pwd, buf, bufsize, &pwd))) {
+			pwd = NULL;
+		}
+	}
+#else
+	pwd = getpwnam(name);
+#endif
 
 	if (pwd != NULL) {
 		if (strcmp( pwd->pw_passwd, "*NP*" ) == 0)
@@ -345,6 +372,10 @@ int _unix_blankpasswd(unsigned int ctrl, const char *name)
 					setreuid( 0, -1 );
 					if(setreuid( -1, pwd->pw_uid ) == -1)
 						/* Will fail elsewhere. */
+#if HAVE_GETPWNAM_R
+						if (buf)
+							free(buf);
+#endif
 						return 0;
 				}
 			}
@@ -384,6 +415,11 @@ int _unix_blankpasswd(unsigned int ctrl, const char *name)
 	if (salt)
 		_pam_delete(salt);
 
+#if HAVE_GETPWNAM_R
+	if (buf)
+		free(buf);
+#endif
+
 	return retval;
 }
 
@@ -394,7 +430,8 @@ int _unix_blankpasswd(unsigned int ctrl, const char *name)
 #include <sys/types.h>
 #include <sys/wait.h>
 
-static int _unix_run_helper_binary(pam_handle_t *pamh, const char *passwd, unsigned int ctrl)
+static int _unix_run_helper_binary(pam_handle_t *pamh, const char *passwd,
+				   unsigned int ctrl, const char *user)
 {
     int retval, child, fds[2];
 
@@ -408,8 +445,8 @@ static int _unix_run_helper_binary(pam_handle_t *pamh, const char *passwd, unsig
     /* fork */
     child = fork();
     if (child == 0) {
-	static char *args[] = { NULL, NULL };
 	static char *envp[] = { NULL };
+	char *args[] = { NULL, NULL, NULL };
 
 	/* XXX - should really tidy up PAM here too */
 
@@ -419,6 +456,8 @@ static int _unix_run_helper_binary(pam_handle_t *pamh, const char *passwd, unsig
 
 	/* exec binary helper */
 	args[0] = x_strdup(CHKPWD_HELPER);
+	args[1] = x_strdup(user);
+
 	execve(CHKPWD_HELPER, args, envp);
 
 	/* should not get here: exit with error */
@@ -530,7 +569,7 @@ int _unix_verify_password(pam_handle_t * pamh, const char *name
 		if (geteuid()) {
 			/* we are not root perhaps this is the reason? Run helper */
 			D(("running helper binary"));
-			retval = _unix_run_helper_binary(pamh, p, ctrl);
+			retval = _unix_run_helper_binary(pamh, p, ctrl, name);
 			if (pwd == NULL && !on(UNIX_AUDIT,ctrl)
 			    && retval != PAM_SUCCESS)
 			{
@@ -567,6 +606,7 @@ int _unix_verify_password(pam_handle_t * pamh, const char *name
 			if (!strncmp(salt, "$1$", 3)) {
 				pp = Goodcrypt_md5(p, salt);
 				if (strcmp(pp, salt) != 0) {
+					_pam_delete(pp);
 					pp = Brokencrypt_md5(p, salt);
 				}
 			} else {
@@ -577,7 +617,16 @@ int _unix_verify_password(pam_handle_t * pamh, const char *name
 			/* the moment of truth -- do we agree with the password? */
 			D(("comparing state of pp[%s] and salt[%s]", pp, salt));
 
-			if (strcmp(pp, salt) == 0) {
+			/*
+			 * Note, we are comparing the bigcrypt of the password with
+			 * the contents of the password field. If the latter was
+			 * encrypted with regular crypt (and not bigcrypt) it will
+			 * have been truncated for storage relative to the output
+			 * of bigcrypt here. As such we need to compare only the
+			 * stored string with the subset of bigcrypt's result.
+			 * Bug 521314: The strncmp comparison is for legacy support.
+			 */
+			if (strncmp(pp, salt, strlen(salt)) == 0) {
 				retval = PAM_SUCCESS;
 			} else {
 				retval = PAM_AUTH_ERR;
@@ -658,7 +707,7 @@ int _unix_verify_password(pam_handle_t * pamh, const char *name
 	if (salt)
 		_pam_delete(salt);
 	if (pp)
-		_pam_overwrite(pp);
+		_pam_delete(pp);
 
 	D(("done [%d].", retval));
 
@@ -679,7 +728,6 @@ int _unix_read_password(pam_handle_t * pamh
 {
 	int authtok_flag;
 	int retval;
-	const char *item;
 	char *token;
 
 	D(("called"));
@@ -701,16 +749,14 @@ int _unix_read_password(pam_handle_t * pamh
 	 */
 
 	if (on(UNIX_TRY_FIRST_PASS, ctrl) || on(UNIX_USE_FIRST_PASS, ctrl)) {
-		retval = pam_get_item(pamh, authtok_flag, (const void **) &item);
+		retval = pam_get_item(pamh, authtok_flag, (const void **) pass);
 		if (retval != PAM_SUCCESS) {
 			/* very strange. */
 			_log_err(LOG_ALERT, pamh
 				 ,"pam_get_item returned error to unix-read-password"
 			    );
 			return retval;
-		} else if (item != NULL) {	/* we have a password! */
-			*pass = item;
-			item = NULL;
+		} else if (*pass != NULL) {	/* we have a password! */
 			return PAM_SUCCESS;
 		} else if (on(UNIX_USE_FIRST_PASS, ctrl)) {
 			return PAM_AUTHTOK_RECOVER_ERR;		/* didn't work */
@@ -809,9 +855,10 @@ int _unix_read_password(pam_handle_t * pamh
 		_pam_delete(token);	/* clean it up */
 		if (retval != PAM_SUCCESS
 		    || (retval = pam_get_item(pamh, authtok_flag
-					      ,(const void **) &item))
+					      ,(const void **) pass))
 		    != PAM_SUCCESS) {
 
+			*pass = NULL;
 			_log_err(LOG_CRIT, pamh, "error manipulating password");
 			return retval;
 
@@ -830,13 +877,10 @@ int _unix_read_password(pam_handle_t * pamh
 			_pam_delete(token);
 			return retval;
 		}
-		item = token;
+		*pass = token;
 		token = NULL;	/* break link to password */
 	}
 
-	*pass = item;
-	item = NULL;		/* break link to password */
-
 	return PAM_SUCCESS;
 }
 
