@@ -1,5 +1,5 @@
 /*
- * $Id: unix_chkpwd.c,v 1.3 2001/02/11 06:33:53 agmorgan Exp $
+ * $Id: unix_chkpwd.c,v 1.8 2002/07/09 04:44:18 agmorgan Exp $
  *
  * This program is designed to run setuid(root) or with sufficient
  * privilege to read all of the unix password databases. It is designed
@@ -15,6 +15,12 @@
 
 #include <security/_pam_aconf.h>
 
+#ifdef MEMORY_DEBUG
+# undef exit
+# undef strdup
+# undef free
+#endif /* MEMORY_DEBUG */
+
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -143,7 +149,16 @@ static int _unix_verify_password(const char *name, const char *p, int opt)
 		}
 	} else {
 		pp = bigcrypt(p, salt);
-		if (strcmp(pp, salt) == 0) {
+		/*
+		 * Note, we are comparing the bigcrypt of the password with
+		 * the contents of the password field. If the latter was
+		 * encrypted with regular crypt (and not bigcrypt) it will
+		 * have been truncated for storage relative to the output
+		 * of bigcrypt here. As such we need to compare only the
+		 * stored string with the subset of bigcrypt's result.
+		 * Bug 521314: the strncmp comparison is for legacy support.
+		 */
+		if (strncmp(pp, salt, strlen(salt)) == 0) {
 			retval = UNIX_PASSED;
 		}
 	}
@@ -155,6 +170,7 @@ static int _unix_verify_password(const char *name, const char *p, int opt)
 		if (pp != NULL) {
 			while (tp && *tp)
 				*tp++ = '\0';
+			free(pp);
 		}
 		pp = tp = NULL;
 	}
@@ -171,9 +187,8 @@ static char *getuidname(uid_t uid)
 	if (pw == NULL)
 		return NULL;
 
-	memset(username, 0, 32);
-	strncpy(username, pw->pw_name, 32);
-	username[31] = '\0';
+	strncpy(username, pw->pw_name, sizeof(username));
+	username[sizeof(username) - 1] = '\0';
 	
 	return username;
 }
@@ -225,7 +240,7 @@ int main(int argc, char *argv[])
 	    }
 	}
 
-	/* read the nollok/nonull option */
+	/* read the nullok/nonull option */
 
 	npass = read(STDIN_FILENO, option, 8);
 
